import logging
from typing import Any
from qua_dashboards.utils.dash_utils import create_input_field
import dash_bootstrap_components as dbc
from dash import html, ALL, no_update, Dash, Input, Output, State, dcc, callback_context, MATCH
from qua_dashboards.core.base_updatable_component import ModifiedFlags


from qua_dashboards.video_mode.tab_controllers.base_tab_controller import (
    BaseTabController,
)
from qua_dashboards.video_mode.data_acquirers.base_data_acquirer import BaseDataAcquirer

logger = logging.getLogger(__name__)

__all__ = ["SettingsTabController"]


class SettingsTabController(BaseTabController):
    """
    Controls the 'Settings' tab in the Video Mode application.
    This tab allows the user to adjust the settings of the measurement, such as the readout power and frequency, or the readout parameter (I/Q/R/Phase)
    """

    _TAB_LABEL = "Settings"
    _TAB_VALUE = "settings-tab"
    _DUMMY_OUT_SUFFIX = "dummy-settings-updates"

    def __init__(
        self,
        data_acquirer: BaseDataAcquirer,
        component_id: str = "settings-tab",
        is_active: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initializes the Settings Tab.
        Args:
            component_id: A unique string identifier for this component instance.
            data_acquirer: The Data Acquirer instance that the measurement uses
            **kwargs: Additional keyword arguments passed to BaseComponent.
        """
        super().__init__(component_id=component_id, is_active=is_active, **kwargs)
        self._data_acquirer_instance = data_acquirer
        logger.info(
            f"Settings Tab '{self.component_id}' initialized with "
            f"Data Acquirer '{self._data_acquirer_instance}'."
        )
        self._inner_loop_functions = self._data_acquirer_instance.inner_functions_dict

    def get_layout(self):
        # ramp_duration_input = create_input_field(
        #     id={
        #         "type": "ramp_duration",
        #         "index": f"{self._data_acquirer_instance.component_id}::ramp_duration",
        #     },
        #     label="Ramp Duration",
        #     value=getattr(self._data_acquirer_instance.inner_loop_action, "ramp_duration", 16),
        #     units="ns",
        #     step=4,
        # )
        pre_measurement_delay_input = create_input_field(
            id={
                "type": "pre_measurement_delay",
                "index": f"{self._data_acquirer_instance.component_id}::pre_measurement_delay",
            },
            label="Pre-Measurement Delay",
            value=getattr(self._data_acquirer_instance.inner_loop_action, "pre_measurement_delay", 0),
            units="ns",
            step=4,
        )

        inner_controls = []
        try:
            inner_controls = (
                self._data_acquirer_instance.inner_loop_action.get_dash_components(
                    include_subcomponents=True
                )
            )
        except: 
            pass
        readout_selector = dbc.Row(
            [
                dbc.Label(
                    "Readouts to acquire", width="auto", className="col-form-label"
                ),
                dbc.Col(
                    dcc.Dropdown(
                        id=self._data_acquirer_instance._get_id("readouts"),
                        options=[
                            {"label": ch, "value": ch}
                            for ch in self._data_acquirer_instance.available_readout_channels.keys()
                        ],
                        value=[
                            ch.name
                            for ch in self._data_acquirer_instance.selected_readout_channels
                        ],
                        multi=True,
                        clearable=False,
                        style={"color": "black"},
                    ),
                    width=True,
                ),
            ],
            className="mb-2 align-items-center",
        )

        inner_loop_functions = ["Go to Point(s)"] + list(self._inner_loop_functions.keys())
        inner_function_selector = dbc.Row([            
            dbc.Label("Inner Loop Function", width = "auto", className = "col-form-label"), 
            dbc.Col(
                dcc.Dropdown(
                    id = self._data_acquirer_instance._get_id("inner-loop-function"), 
                    options = [
                        {"label": fn_name, "value": fn_name} for fn_name in inner_loop_functions
                    ], 
                    value = None, 
                    multi= False,
                    clearable = True, 
                    style = {"color": "black"}, 
                ), 
                width = True,
            ),
            ], 
            className = "mb-2 align-items-center"
        )
        point_sequence_section = self._build_point_sequence_section()
        inner_loop_section = html.Div(
            [
                html.Small("Inner Loop Action", className="text-light mb-2", style={"display": "block"}),
                inner_function_selector,
                point_sequence_section,
            ],
            style={
                "outline": "2px solid #fff",
                "borderRadius": "8px",
                "padding": "12px",
                "marginBottom": "12px",
            },
        )

        post_processing_fn_selector = dbc.Row([            
            dbc.Label("Post-Processing Function", width = "auto", className = "col-form-label"), 
            dbc.Col(
                dcc.Dropdown(
                    id = self._data_acquirer_instance._get_id("post-processing-function"), 
                    options = [
                        {"label": fn_name, "value": fn_name} for (fn_name) in self._data_acquirer_instance.post_processing_functions.keys()
                    ], 
                    value = "Raw_data", 
                    multi= False,
                    clearable = False, 
                    style = {"color": "black"}, 
                ), 
                width = True,
            ),
            ], 
            className = "mb-2 align-items-center"
        )

        scan_mode_selector = dbc.Row([            
            dbc.Label("2D Scan Mode", width = "auto", className = "col-form-label"), 
            dbc.Col(
                dcc.Dropdown(
                    id = self._data_acquirer_instance._get_id("scan-mode-selection"), 
                    options = [
                        {"label": scan_mode, "value": scan_mode} for (scan_mode) in self._data_acquirer_instance.scan_modes.keys()
                    ], 
                    value = self._data_acquirer_instance.current_scan_mode, 
                    multi= False,
                    clearable = False, 
                    style = {"color": "black"}, 
                ), 
                width = True,
            ),
            ], 
            className = "mb-2 align-items-center"
        )

        result_type_selector = dbc.Row(
            [
                dbc.Label("Result Type", width="auto", className="col-form-label"),
                dbc.Col(
                    dbc.Select(
                        id={
                            "type": "select",
                            "index": f"{self._data_acquirer_instance.component_id}::result-type",
                        },
                        options=[
                            {"label": rt, "value": rt}
                            for rt in self._data_acquirer_instance.result_types
                        ],
                        value=self._data_acquirer_instance.result_type,
                        style={"width": "150px"},
                    ),
                    width=True,
                ),
            ],
            className="mb-2 align-items-center",
        )

        display_components = [
            readout_selector, 
            result_type_selector, 
            post_processing_fn_selector,
            scan_mode_selector, 
        ]
        if hasattr(self._data_acquirer_instance.inner_loop_action, "pre_measurement_delay"): 
            display_components.append(pre_measurement_delay_input)

        return dbc.Card(
            dbc.CardBody(
                [
                    html.H5("Settings", className="text-light"),
                    *display_components,
                    inner_loop_section,
                    *inner_controls,
                    html.Div(
                        id=self._get_id(self._DUMMY_OUT_SUFFIX),
                        style={"display": "none"},
                    ),
                ]
            ),
            color="dark",
            inverse=True,
            className="tab-card-dark",
        )

    def on_tab_activated(self):
        logger.info(f"SettingsTabController '{self.component_id}' activated.")
        self._data_acquirer_instance.stop_acquisition()
        from qua_dashboards.video_mode.video_mode_component import VideoModeComponent

        super().on_tab_activated()
        return {
            VideoModeComponent._MAIN_STATUS_ALERT_ID_SUFFIX: html.Span(
                "STOPPED", style={"display": "none"}
            )
        }

    def on_tab_deactivated(self):
        logger.info(f"SettingsTabController '{self.component_id}' deactivated.")
        return super().on_tab_deactivated()

    def register_callbacks(self, app: Dash, **kwargs):
        acq = self._data_acquirer_instance
        dummy_out = self._get_id(self._DUMMY_OUT_SUFFIX)
        main_status_id = kwargs["orchestrator_stores"]["main-status-alert"]

        @app.callback(
            Output(main_status_id, "children", allow_duplicate=True),
            Output(dummy_out, "children", allow_duplicate = True),
            Input({"type": "comp-inner-loop", "index": ALL}, "value"),
            State({"type": "comp-inner-loop", "index": ALL}, "id"),
            Input({"type": "select", "index": ALL}, "value"),
            State({"type": "select", "index": ALL}, "id"),
            Input({"type": "ramp_duration", "index": ALL}, "value"),
            State({"type": "ramp_duration", "index": ALL}, "id"),
            Input({"type": "pre_measurement_delay", "index": ALL}, "value"),
            State({"type": "pre_measurement_delay", "index": ALL}, "id"),
            prevent_initial_call=True,
        )
        def _apply_settings(
            inner_vals, inner_ids, select_vals, select_ids, ramp_vals, ramp_ids, pre_meas_delay_vals, pre_meas_delay_ids
        ):
            """
            Collect changes from the Settings tab and forward them to the acquirer.
            - Inner loop controls use ids like {'type': 'comp-inner-loop', 'index': 'readout_duration'}
            - Result type select uses id like {'type': 'select', 'index': 'opx-data-acquirer::result-type'}
            """
            params_to_update = {}
            if inner_vals and inner_ids:
                for v, idd in zip(inner_vals, inner_ids):
                    if not isinstance(idd, dict):
                        continue
                    param = idd.get("index")
                    if param is None:
                        continue
                    params_to_update.setdefault(
                        acq.inner_loop_action.component_id, {}
                    )[param] = v
            if select_vals and select_ids:
                idx = select_ids[0].get("index")
                comp_id, param = idx.split("::", 1)
                params_to_update.setdefault(comp_id, {})[param] = select_vals[0]

            if ramp_vals and ramp_ids:
                idx = ramp_ids[0].get("index")
                comp_id, param = idx.split("::", 1)
                if ramp_vals[0] is not None: 
                    ramp_vals[0] = int(ramp_vals[0])
                params_to_update.setdefault(comp_id, {})[param] = ramp_vals[0]
            if pre_meas_delay_vals and pre_meas_delay_ids:
                idx = pre_meas_delay_ids[0].get("index")
                comp_id, param = idx.split("::", 1)
                if pre_meas_delay_vals[0] is not None: 
                    pre_meas_delay_vals[0] = int(pre_meas_delay_vals[0])
                params_to_update.setdefault(comp_id, {})[param] = pre_meas_delay_vals[0]
            if getattr(self, "_last_total_pixel_duration_ns", None) is not None:
                params_to_update.setdefault(acq.component_id, {})["total_pixel_duration_ns"] = int(self._last_total_pixel_duration_ns)

            if not params_to_update:
                return no_update, no_update
            try: 
                acq.update_parameters(params_to_update)
                return "", no_update
            except Exception as e:  
                import dash_bootstrap_components as dbc
                return dbc.Alert(str(e), color="danger", dismissable=True), no_update
        
        @app.callback(
            Output(dummy_out, "children", allow_duplicate = True), 
            Input(self._data_acquirer_instance._get_id("post-processing-function"), "value"), 
            prevent_initial_call = True,
        ) 
        def _alter_post_processing_function(fn_name): 
            self._data_acquirer_instance.selected_function = self._data_acquirer_instance.post_processing_functions[fn_name]
            logger.info(f"Post Processing Function Changed to {fn_name}")
            return no_update
        
        @app.callback(
            Output(dummy_out, "children", allow_duplicate = True), 
            Input(self._data_acquirer_instance._get_id("scan-mode-selection"), "value"), 
            prevent_initial_call = True,
        ) 
        def _alter_scan_mode(name): 
            self._data_acquirer_instance.set_scan_mode(name)
            logger.info(f"Scan Mode Changed to {name}")
            return no_update
        
        @app.callback(
            Output(f"{self.component_id}-point-sequence-section", "style"), 
            Input(acq._get_id("inner-loop-function"), "value"), 
            prevent_initial_call=True,
        )
        def _toggle_point_section(fn_name): 
            if fn_name == "Go to Point(s)": 
                return {"display": "block"}
            elif fn_name in self._inner_loop_functions: 
                user_function = self._inner_loop_functions[fn_name]
                def loop_action(inner_loop_self): 
                    user_function()
                if hasattr(acq.inner_loop_action, "loop_action"):
                    acq.inner_loop_action.loop_action = loop_action
                acq._compilation_flags |= ModifiedFlags.PROGRAM_MODIFIED
                return {"display": "none"}
            
            else: 
                def loop_action(inner_loop_self): 
                    pass
                def pre_loop_action(inner_loop_self): 
                    pass
                if hasattr(acq.inner_loop_action, "loop_action"):
                    acq.inner_loop_action.loop_action = loop_action
                if hasattr(acq.inner_loop_action, "pre_loop_action"):
                    acq.inner_loop_action.pre_loop_action = pre_loop_action
                acq._compilation_flags |= ModifiedFlags.PROGRAM_MODIFIED
                return {"display": "none"}
        
        @app.callback(
            Output(f"{self.component_id}-point-rows-container", "children"),
            Output(main_status_id, "children", allow_duplicate=True),
            Input(f"{self.component_id}-add-point-btn", "n_clicks"),
            Input({"type": "remove-point-btn", "index": ALL}, "n_clicks"),
            Input({"type": "point-timing", "index": ALL}, "value"),
            State({"type": "point-select", "index": ALL}, "value"),
            State({"type": "point-duration", "index": ALL}, "value"),
            State(f"{self.component_id}-xy-duration", "value"),
            State({"type": "point-ramp", "index": ALL}, "value"),
            State(f"{self.component_id}-xy-ramp", "value"),
            prevent_initial_call=True,
        )
        def _manage_point_rows(add_clicks, remove_clicks, timings, point_names, durations, xy_duration, point_ramps, xy_ramp):
            ctx = callback_context
            if not ctx.triggered:
                return no_update, no_update
            
            trigger_id = ctx.triggered[0]["prop_id"].split(".")[0]
            gate_set = acq.gate_set
            
            if not hasattr(gate_set, 'macros') or not gate_set.get_macros():
                if trigger_id == f"{self.component_id}-add-point-btn":
                    return no_update, dbc.Alert("No points defined in the GateSet. Identify some in the Annotation Tab first.", color="warning", dismissable=True, duration=4000)
                return no_update, no_update
            
            available_points = [{"label": name, "value": name} for name in gate_set.get_macros().keys()]
            rows_data = list(zip(timings, point_names, durations, point_ramps))
            
            if trigger_id == f"{self.component_id}-add-point-btn":
                rows_data.append(("pre", available_points[0]["value"] if available_points else None, 1000, 16))
            else:
                try:
                    trigger_dict = eval(trigger_id)
                    if trigger_dict.get("type") == "remove-point-btn":
                        remove_idx = trigger_dict["index"]
                        rows_data = [r for i, r in enumerate(rows_data) if i != remove_idx]
                except:
                    return no_update, no_update
            
            pre_rows = [(t, p, int(d), int(r)) for t, p, d, r in rows_data if t == "pre"]
            post_rows = [(t, p, int(d), int(r)) for t, p, d, r in rows_data if t == "post"]

            pre_points = [(p, int(d), int(r)) for t, p, d, r in rows_data if t == "pre" and p and d]
            post_points = [(p, int(d), int(r)) for t, p, d, r in rows_data if t == "post" and p and d]
            
            if hasattr(acq.inner_loop_action, "point_duration"):
                acq.inner_loop_action.point_duration = int(xy_duration) if xy_duration else 0
            if hasattr(acq.inner_loop_action, "ramp_duration"):
                acq.inner_loop_action.ramp_duration = int(xy_ramp) if xy_ramp else 16
            
            def pre_loop_action(inner_loop_self):
                for point, duration, ramp_duration in pre_points:
                    inner_loop_self.voltage_sequence.ramp_to_point(
                        point, duration=duration, ramp_duration=ramp_duration
                    )
            
            def loop_action(inner_loop_self):
                for point, duration, ramp_duration in post_points:
                    inner_loop_self.voltage_sequence.ramp_to_point(
                        point, duration=duration, ramp_duration=ramp_duration
                    )
            
            if hasattr(acq.inner_loop_action, "loop_action"):
                acq.inner_loop_action.loop_action = loop_action
            if hasattr(acq.inner_loop_action, "pre_loop_action"):
                acq.inner_loop_action.pre_loop_action = pre_loop_action
            acq._compilation_flags |= ModifiedFlags.PROGRAM_MODIFIED
            logger.info(f"Point sequence updated from row manager - Pre: {pre_points}, Post: {post_points}")
            
            result = [self._build_point_header_row()]
            for i, (t, p, d, r) in enumerate(pre_rows):
                result.append(self._build_point_row(i, available_points, t, p, d, r))
            
            result.append(self._build_xy_row(xy_duration or 0, xy_ramp or 16))
            
            for i, (t, p, d, r) in enumerate(post_rows):
                result.append(self._build_point_row(len(pre_rows) + i, available_points, t, p, d, r))
            
            return result, no_update
        
        @app.callback(
            Output({"type": "point-duration", "index": MATCH}, "value"),
            Input({"type": "point-select", "index": MATCH}, "value"),
            prevent_initial_call=True,
        )
        def _update_duration_from_point(point_name):
            if not point_name:
                return no_update
            
            macros = acq.gate_set.get_macros()
            if point_name in macros:
                return macros[point_name].duration
            return no_update
                
        @app.callback(
            Output(dummy_out, "children", allow_duplicate=True), 
            Input({"type": "point-select", "index": ALL}, "value"),
            Input({"type": "point-duration", "index": ALL}, "value"), 
            Input({"type": "point-timing", "index": ALL}, "value"),
            Input(f"{self.component_id}-xy-duration", "value"),
            Input({"type": "point-ramp", "index": ALL}, "value"),
            Input(f"{self.component_id}-xy-ramp", "value"),
            prevent_initial_call = True,
        )
        def _update_point_sequence(point_names, durations, timings, xy_duration, point_ramps, xy_ramp): 
            pre_points = []
            post_points = []

            for point, duration, timing, ramp in list(zip(point_names, durations, timings, point_ramps)): 
                if point and duration: 
                    ramp_val = int(ramp or 16)
                    if timing == "pre": 
                        pre_points.append((point, int(duration), ramp_val))
                    else: 
                        post_points.append((point, int(duration), ramp_val))
            if hasattr(acq.inner_loop_action, "point_duration"): 
                acq.inner_loop_action.point_duration = int(xy_duration or 0)
            if hasattr(acq.inner_loop_action, "ramp_duration"): 
                acq.inner_loop_action.ramp_duration = int(xy_ramp or 16)
            def pre_loop_action(inner_loop_self): 
                for point, duration, ramp in pre_points: 
                    inner_loop_self.voltage_sequence.ramp_to_point(point, duration = duration, ramp_duration = ramp)
            def loop_action(inner_loop_self): 
                for point, duration, ramp in post_points: 
                    inner_loop_self.voltage_sequence.ramp_to_point(point, duration = duration, ramp_duration = ramp)

            if hasattr(acq.inner_loop_action, "loop_action"):
                acq.inner_loop_action.loop_action = loop_action
            if hasattr(acq.inner_loop_action, "pre_loop_action"):
                acq.inner_loop_action.pre_loop_action = pre_loop_action
            logger.info(f"Point sequence updated - Pre: {pre_points}, Post: {post_points}")
            acq._compilation_flags |= ModifiedFlags.PROGRAM_MODIFIED
            return no_update
        

        @app.callback(
            Output(f"{self.component_id}-timeline-display", "children"),
            Input({"type": "point-select", "index": ALL}, "value"),
            Input({"type": "point-duration", "index": ALL}, "value"),
            Input({"type": "point-timing", "index": ALL}, "value"),
            Input(f"{self.component_id}-xy-duration", "value"),
            Input({"type": "pre_measurement_delay", "index": ALL}, "value"),
            Input({"type": "comp-inner-loop", "index": ALL}, "value"),
            State({"type": "comp-inner-loop", "index": ALL}, "id"),
            Input({"type": "point-ramp", "index": ALL}, "value"),
            Input(f"{self.component_id}-xy-ramp", "value"),
            prevent_initial_call=False,
        )
        def _update_timeline(
            point_names, point_durations, timings, 
            xy_dur, pre_meas_vals,
            inner_vals, inner_ids, 
            point_ramps, xy_ramp,
        ):
            inner_loop = acq.inner_loop_action
            pre_measurement_delay = pre_meas_vals[0] if pre_meas_vals else 0
            readout_duration = 0
            if hasattr(inner_loop, "_pulse_for"): 
                readout_duration = max(
                    inner_loop._pulse_for(ch).length 
                    for ch in inner_loop.selected_readout_channels
                ) if inner_loop.selected_readout_channels else 0
            time_of_flight = max(ch.time_of_flight for ch in inner_loop.selected_readout_channels)

            pre_points = []
            post_points = []
            
            for point, duration, timing, ramp in zip(point_names, point_durations, timings, point_ramps):
                if point and duration:
                    ramp_val = int(ramp or 16)
                    if timing == "pre":
                        pre_points.append((point, int(duration), ramp_val))
                    else:
                        post_points.append((point, int(duration), ramp_val))
            
            return self._build_timeline_visuals(
                        pre_points=pre_points,
                        xy_duration=xy_dur or 0,
                        xy_ramp = int(xy_ramp or 16),
                        post_points=post_points,
                        pre_measurement_delay=pre_measurement_delay,
                        readout_duration=readout_duration,
                        time_of_flight=time_of_flight
                    )
    def _build_point_header_row(self) -> dbc.Row:
        return dbc.Row([
            dbc.Col(html.Small("Timing", className="text-light"), width=2),
            dbc.Col(html.Small("Point", className="text-light"), width=5),
            dbc.Col(html.Small("Ramp Dur (ns)", className="text-light"), width=2),
            dbc.Col(html.Small("Point Dur (ns)", className="text-light"), width=3),
        ], className="mb-1 g-1 align-items-center flex-nowrap pe-1")
        
    def _build_xy_row(self, xy_duration = 0, xy_ramp = 16): 
        return dbc.Row(
            [
                dbc.Col(html.Span(""), width=2),
                dbc.Col(html.Span("(x,y)", style={"fontWeight": "bold"}), width=5),
                dbc.Col(
                    dbc.Input(
                        id=f"{self.component_id}-xy-ramp",
                        type="number",
                        value=xy_ramp,
                        min=4,
                        step=4,
                        size="sm",
                        style={"width": "70px"},
                    ),
                    width=2,
                ),
                dbc.Col(
                    dbc.Input(
                        id=f"{self.component_id}-xy-duration",
                        type="number",
                        value=xy_duration,
                        min=0,
                        step=4,
                        size="sm",
                        style={"width": "80px"},
                    ),
                    width="auto", 
                    className="px-0 me-1"
                ),
            ],
            className="mb-1 g-1 align-items-center flex-nowrap pe-1",
            style={"padding": "0px"},
        )

    def _build_point_sequence_section(self): 
        """
        Build a collapsible section for point sequence configuration
        """

        inner_loop = self._data_acquirer_instance.inner_loop_action
        xy_ramp = getattr(inner_loop, "ramp_duration", 16)
        pre_measurement_delay = getattr(inner_loop, "pre_measurement_delay", 0)
        readout_duration = 0
        if hasattr(inner_loop, "_pulse_for"): 
            readout_duration = max(
                inner_loop._pulse_for(ch).length 
                for ch in inner_loop.selected_readout_channels
            ) if inner_loop.selected_readout_channels else 0
        time_of_flight = max(ch.time_of_flight for ch in inner_loop.selected_readout_channels)

        xy_duration=int(getattr(inner_loop, "point_duration", 0) or 0)
        initial_timeline = self._build_timeline_visuals(
            pre_points=[],
            xy_duration=xy_duration,
            post_points=[],
            xy_ramp = xy_ramp,
            pre_measurement_delay=pre_measurement_delay,
            readout_duration=readout_duration,
            time_of_flight = time_of_flight
        )

        return html.Div([
                html.Div([
                    html.Div(id=f"{self.component_id}-point-rows-container", children=[self._build_point_header_row(),self._build_xy_row(xy_duration, xy_ramp)]),
                            dbc.Button("+ Add Point", id=f"{self.component_id}-add-point-btn", size="sm", className="mt-2"),
                        ],
                    id=f"{self.component_id}-point-sequence-section",
                    style={"display": "none"},
                    ), 
                    html.Div(
                        id=f"{self.component_id}-timeline-display",
                        children=initial_timeline,
                        style={"marginTop": "12px"},
                    ),
                ])

    def _build_point_row(self, index: int, available_points: list, timing_value = "pre", point_value = None, duration_value = None, ramp_value=None) -> dbc.Row: 
        macros = self._data_acquirer_instance.gate_set.get_macros()
        if point_value is None: 
            point_value = available_points[0]["value"] if available_points else None

        if duration_value is None:
            duration_value = 1000 
            if point_value and duration_value in macros:
                macro = macros[point_value]
                duration_value = macro.duration
        if ramp_value is None: 
            ramp_value = 16

        row = dbc.Row(
            [
                dbc.Col(
                    dbc.Select(
                        id={"type": "point-timing", "index": index},
                        options=[
                            {"label": "Before (x,y)", "value": "pre"},
                            {"label": "After (x,y)", "value": "post"},
                        ],
                        value=timing_value,
                        size="sm",
                        style={"width": "100%"},
                    ),
                    width=2,
                ),
                dbc.Col(
                    dbc.Select(
                        id = {"type": "point-select", "index": index}, 
                        options = available_points, 
                        value = point_value,
                        # clearable = True, 
                        size = "sm",
                        style = {"color": "black", "width": "100%"}
                    ), 
                    width = 5,
                ), 
                dbc.Col(
                    dbc.Input(
                        id={"type": "point-ramp", "index": index},
                        type="number",
                        value=ramp_value,
                        min=16,
                        step=4,
                        size="sm",
                        style={"width": "70px"},
                    ),
                    width=2,
                ),
                dbc.Col(
                    dbc.Input(
                        id={"type": "point-duration", "index": index},
                        type="number",
                        value=duration_value,
                        min=16,
                        step=4,
                        size="sm",
                        style={"width": "80px"},
                    ),
                    width=3,
                ),
                dbc.Col(
                    dbc.Button(
                        "X",
                        id={"type": "remove-point-btn", "index": index},
                        color="danger",
                        size="sm",
                        className="px-1",
                        style={"marginLeft": "-6px"}
                    ),
                    width="auto",
                    className="ps-0",
                ),
            ], 
            className="mb-1 g-1 align-items-center flex-nowrap pe-1"
        )

        return row

    def _build_timeline_visuals(
        self, 
        pre_points, 
        xy_duration, 
        xy_ramp,
        post_points, 
        pre_measurement_delay, 
        readout_duration, 
        time_of_flight,
    ) -> None: 
        """
        A function which builds the visual representation of the pixel timeline, based on the point sequence. Prints the total point duration
        """

        segments = []
        total_duration = 0

        for name, duration, ramp in pre_points: 
            segments.append({"label": name, "duration": duration, "type": "pre", "ramp": ramp})
            total_duration = total_duration + duration + ramp
        
        segments.append({"label": "(x,y)", "duration": xy_duration, "type": "xy", "ramp": xy_ramp})
        total_duration = total_duration + xy_duration + xy_ramp

        for name, duration, ramp in post_points: 
            segments.append({"label": name, "duration": duration, "type": "post", "ramp": ramp})
            total_duration = total_duration + duration + ramp

        segments.append({"label": "PMD", "duration": pre_measurement_delay, "type": "pmw", "ramp": 0})
        segments.append({"label": "Readout", "duration": readout_duration, "type": "readout", "ramp": 0})
        total_duration += pre_measurement_delay + readout_duration

        colours = {"pre": "#6c9bd1", "xy": "#f0ad4e", "post": "#5cb85c", "pmw": "#6f5e96", "readout": "#d9534f", "ramp": "#eaff00"}
        blocks = []
        block_height = "40px"
        for seg in segments: 
            if seg["ramp"] > 0: 
                blocks.append(html.Div(
                    f"Ramp", 
                    style = {
                        "width": f"{40}px",
                        "backgroundColor": colours["ramp"],
                        "textAlign": "center",
                        "lineHeight": block_height,
                        "height": block_height,
                        "fontSize": "10px",
                        "color": "black",
                    }
                ))
            blocks.append(html.Div(
                f"{seg['label']} {seg['duration']}ns", 
                style={
                    "width": f"{100}px",
                    "backgroundColor": colours[seg["type"]],
                    "textAlign": "center",
                    "fontSize": "10px",
                    "height": block_height,
                    "lineHeight": block_height,
                    "color": "black",
                    "overflow": "hidden",
                }
            ))
        readout_overhead = (time_of_flight + 52)
        blocks.append(
            html.Div(
                f"{readout_overhead}ns OH", 
                style={
                    "width": f"{50}px",
                    "backgroundColor": "#566161",
                    "textAlign": "center",
                    "fontSize": "10px",
                    "height": block_height,
                    "lineHeight": block_height,
                    "color": "black",
                    "overflow": "hidden",
                }
            )
        )
        total_duration = total_duration + readout_overhead
        self._last_total_pixel_duration_ns = int(total_duration)
        return html.Div([
            html.Div(blocks, style={"display": "flex", "alignItems": "center", "marginBottom": "8px", "overflowX": "auto",}),
            html.Div(f"Total pixel duration: {total_duration} ns", style={"fontSize": "12px", "color": "#aaa"}),
        ])



import numpy as np
from typing import Sequence, Any, List, Dict, Optional

from qua_dashboards.video_mode.inner_loop_actions.simulators.base_simulator import BaseSimulator

__all__ = ["QarraySimulator"]

class QarraySimulator(BaseSimulator): 
    """
    Simulator using Qarray to simulate a QD structure.  
    """
    def __init__(
        self, *, 
        gate_set, 
        dc_set,
        model, 
        sensor_gate_names: Sequence[str],
        base_point: dict[str, float] | None = None, 
        voltage_scale: float = 1e3,
    ):
        super().__init__(gate_set=gate_set)
        self.model = model
        self.base_point = dict(base_point or {})

        self.sensor_gate_name = sensor_gate_names[0]
        self.sensor_gate_names = tuple(sensor_gate_names)
        self.sensor_index_by_name = {name: i for i, name in enumerate(self.sensor_gate_names)}
        self.dc_set = dc_set
        self.voltage_scale = voltage_scale

        self.qarray_gate_order = self.infer_qarray_gate_order()
        if dc_set is not None:
            _ = dc_set.all_current_voltages


    def infer_qarray_gate_order(self) -> List[str]: 
        gate_set = self.gate_set
        if hasattr(gate_set, "layers") and gate_set.layers: 
            tg = gate_set.layers[0].target_gates
            return tg
        else: 
            return sorted(gate_set.channels.keys())
    
    def point_to_vg(self, point: Dict[str, float]) -> np.ndarray: 
        physical_voltages = self.gate_set.resolve_voltages(point, allow_extra_entries=True)
        return np.array([float(physical_voltages.get(name, 0.0)) for name in self.qarray_gate_order], dtype = float)
    
    # def get_physical_grid(self, x_axis_name, y_axis_name, x_vals, y_vals): 
    #     """x_vals and y_vals are 1d sweeps"""
    #     base = dict(self.dc_set._current_levels)
    #     base.update(self.base_point)

    #     x_vals = np.asarray(x_vals, float)
    #     y_vals = np.asarray(y_vals, float)
    #     x_vals = x_vals - x_vals[len(x_vals)//2]
    #     y_vals = y_vals - y_vals[len(y_vals)//2]

    #     n_gates = len(self.qarray_gate_order)

    #     grids = [np.zeros((len(y_vals), len(x_vals)), float) for _ in range(n_gates)]
    #     for iy, dy in enumerate(y_vals):
    #         for ix, dx in enumerate(x_vals):
    #             point = dict(base)
    #             point[x_axis_name] = point.get(x_axis_name, 0.0) + dx
    #             point[y_axis_name] = point.get(y_axis_name, 0.0) + dy
    #             phys = self.gate_set.resolve_voltages(
    #                 point,
    #                 allow_extra_entries=True,
    #             )
    #             for gi, gate_name in enumerate(self.qarray_gate_order):
    #                 grids[gi][iy, ix] = float(phys.get(gate_name, 0.0))

    #     return grids

    def get_physical_grid(self, x_axis_name, y_axis_name, x_vals, y_vals):
        """Vectorized version - much faster for large grids."""
        # base = dict(
        #     self.dc_set._current_levels
        #     if self.dc_set is not None
        #     else zip(self.qarray_gate_order, np.zeros(len(self.qarray_gate_order)))
        # )
        if self.dc_set is not None:
            base = dict(self.dc_set._current_levels)
        else:
            base = dict(self.base_point)
        
        print(base)
        print(self.qarray_gate_order)
        
        x_vals = np.asarray(x_vals, float)
        y_vals = np.asarray(y_vals, float)
        x_vals = x_vals - x_vals[len(x_vals)//2]
        y_vals = y_vals - y_vals[len(y_vals)//2]
        
        base_phys = {name: base.get(name, 0.0) for name in self.qarray_gate_order}
        
        delta_x = self.gate_set.resolve_voltages(
            {**{k: 0.0 for k in base}, x_axis_name: 1.0}, 
            allow_extra_entries=True
        )
        delta_y = self.gate_set.resolve_voltages(
            {**{k: 0.0 for k in base}, y_axis_name: 1.0}, 
            allow_extra_entries=True
        )
        
        n_gates = len(self.qarray_gate_order)
        base_vec = np.array([base_phys.get(g, 0.0) for g in self.qarray_gate_order])
        dx_vec = np.array([delta_x.get(g, 0.0) for g in self.qarray_gate_order])
        dy_vec = np.array([delta_y.get(g, 0.0) for g in self.qarray_gate_order])
        X, Y = np.meshgrid(x_vals, y_vals)
        grids = (base_vec[None, None, :] 
                + X[:, :, None] * dx_vec[None, None, :] 
                + Y[:, :, None] * dy_vec[None, None, :])
        
        return [grids[:, :, i] for i in range(n_gates)]
    
    def grids_to_vg(self, grids): 
        grids = [np.asarray(g, float) for g in grids]
        vg = np.stack(grids, axis=-1)
        return vg


    # def measure_data(
    #     self,
    #     x_axis_name: str,
    #     y_axis_name: str,
    #     x_vals: Sequence[float],
    #     y_vals: Sequence[float],
    #     n_readout_channels: int,
    # ):
    #     """
    #     Run the qarray simulation and return I/Q data.
        
    #     Returns:
    #         I: array of shape (n_readout_channels, y_pts, x_pts)
    #         Q: array of shape (n_readout_channels, y_pts, x_pts)
    #     """
        
    #     grids = self.get_physical_grid(x_axis_name, y_axis_name, x_vals, y_vals)
    #     y_pts, x_pts = grids[0].shape
        
    #     slices = []
    #     for ix in range(x_pts):
    #         vg_slice = np.column_stack([g[:, ix] for g in grids])
    #         vg_slice_mv = vg_slice * self.voltage_scale 
            
    #         raw = self.model.charge_sensor_open(vg_slice_mv)
            
    #         if isinstance(raw, tuple):
    #             raw = raw[0]
    #         slices.append(np.asarray(raw).squeeze())
        

    #     zs = np.asarray(slices)

    #     z = zs.T
        
    #     if z.ndim == 2:
    #         z = z[np.newaxis, ...]
        
    #     n_sensors = z.shape[0]
        
    #     I_list = []
    #     Q_list = []
    #     for ch_idx in range(n_readout_channels):
    #         sensor_idx = ch_idx % n_sensors
    #         signal = z[sensor_idx]
            
    #         I_ch = signal
    #         Q_ch = np.zeros((y_pts, x_pts), dtype=signal.dtype)
            
    #         I_list.append(I_ch)
    #         Q_list.append(Q_ch)
        
    #     return np.stack(I_list, axis=0), np.stack(Q_list, axis=0)
    


    def measure_data(
        self,
        x_axis_name: str,
        y_axis_name: str,
        x_vals: Sequence[float],
        y_vals: Sequence[float],
        n_readout_channels: int,
    ):
        grids = self.get_physical_grid(x_axis_name, y_axis_name, x_vals, y_vals)
        y_pts, x_pts = grids[0].shape
        
        vg = np.stack(grids, axis=-1).reshape(-1, len(grids)) * self.voltage_scale
        
        raw = self.model.charge_sensor_open(vg)  # Model has NO noise
        if isinstance(raw, tuple):
            raw = raw[0]
        raw = np.asarray(raw)
        
        if raw.ndim == 1:
            raw = raw[:, np.newaxis]
        
        n_sensors = raw.shape[-1]
        z = raw.reshape(y_pts, x_pts, n_sensors)
        z = np.moveaxis(z, -1, 0)
        
        # Add noise HERE - no sequential artifacts
        z = z + np.random.normal(0, 0.01, z.shape)
        
        I = np.stack([z[ch_idx % n_sensors] for ch_idx in range(n_readout_channels)])
        Q = np.zeros_like(I)
        
        return I, Q


import dash
from qua_dashboards.logging_config import logger
from qua_dashboards.data_dashboard.data_dashboard_app import DataDashboardApp


logger.info("Registering page data-dashboard")


dash.register_page(__name__)


dashboard_app = DataDashboardApp(app=dash.get_app())

layout = dashboard_app.layout



import logging
import threading
import uuid
from typing import Any, Dict, Optional

logger = logging.getLogger(__name__)

# Simple in-memory cache for single-process Dash applications.
# For multi-process (e.g., Gunicorn with multiple workers), an external
# cache like Redis or a file-system based cache would be necessary.
_DATA_CACHE: Dict[str, Any] = {}
_DATA_VERSIONS: Dict[str, str] = {}  # Stores current version UUID for a key
_CACHE_LOCK: threading.Lock = threading.Lock()

# Define common keys for data stored in the registry
LIVE_DATA_KEY = "live_data"
STATIC_DATA_KEY = "static_data"


def set_data(key: str, data_object: Any) -> str:
    """
    Stores a data object in the registry and generates a new version ID.

    Args:
        key: The key under which to store the data object.
        data_object: The Python object to store.

    Returns:
        The new version UUID string for the stored data.
    """
    new_version = str(uuid.uuid4())
    with _CACHE_LOCK:
        _DATA_CACHE[key] = data_object
        _DATA_VERSIONS[key] = new_version
        logger.debug(
            f"Data set in registry for key '{key}' with new version '{new_version}'"
        )
    return new_version


def get_data(key: str) -> Optional[Any]:
    """
    Retrieves a data object from the registry.

    Args:
        key: The key of the data object to retrieve.

    Returns:
        The stored Python object, or None if the key is not found.
    """
    with _CACHE_LOCK:
        data_object = _DATA_CACHE.get(key)
        if data_object is None:
            logger.warning(f"Data not found in registry for key '{key}'")
        else:
            logger.debug(f"Data retrieved from registry for key '{key}'")
        return data_object


def get_current_version(key: str) -> Optional[str]:
    """
    Retrieves the current version ID for a data key.

    Args:
        key: The key for which to get the version.

    Returns:
        The current version UUID string, or None if the key is not found.
    """
    with _CACHE_LOCK:
        return _DATA_VERSIONS.get(key)



from .data_utils import *
from .dash_utils import *
from .general_utils import *
from .log_utils import *
from .basic_parameter import *
from .example_utils import *

__all__ = [
    *data_utils.__all__,
    *dash_utils.__all__,
    *general_utils.__all__,
    *log_utils.__all__,
    *basic_parameter.__all__,
    *example_utils.__all__,
]



# from time import sleep
# import xarray as xr
# import numpy as np
# from dash.testing.composite import DashComposite

# from qua_dashboards.data_dashboard.data_dashboard_app import DataDashboardApp
# from qua_dashboards.utils.data_utils import send_data_to_dash


# def test_002_data_dashboard_basic_types(dash_duo: DashComposite):
#     data_dashboard_app = DataDashboardApp(include_title=True)
#     dash_duo.start_server(data_dashboard_app.app)

#     dash_duo.wait_for_text_to_equal("#title", "Data Dashboard", timeout=4)
#     assert dash_duo.get_logs() == [], "browser console should contain no error"

#     data = {
#         "text": "Hello, world!",
#         "number": 123,
#         "float": 1.23,
#         "boolean": True,
#         "list": [1, 2, 3],
#         "dict": {"a": 1, "b": 2},
#     }
#     send_data_to_dash(
#         data,
#         url="http://localhost:58050/data-dashboard",
#     )
#     dash_duo.wait_for_element("#data-entry-text", timeout=4)

#     for key, value in data.items():
#         element = dash_duo.find_element(f"#data-entry-{key}")
#         assert element.text == f"{key}:  {value}"

#     assert dash_duo.get_logs() == [], "browser console should contain no error"


# def test_003_data_dashboard_data_array(dash_duo: DashComposite):
#     data_dashboard_app = DataDashboardApp(include_title=True)
#     dash_duo.start_server(data_dashboard_app.app)

#     dash_duo.wait_for_text_to_equal("#title", "Data Dashboard", timeout=4)
#     assert dash_duo.get_logs() == [], "browser console should contain no error"

#     data = {
#         "text": "Hello, world!",
#         "array": xr.DataArray(np.random.rand(10, 10), name=f"my_arr_1"),
#         "array2": xr.DataArray(np.random.rand(100, 100), name=f"my_arr_2"),
#     }
#     send_data_to_dash(data, url="http://localhost:58050/data-dashboard")
#     dash_duo.wait_for_element("#data-entry-text", timeout=4)

#     assert dash_duo.find_element("#data-entry-text").text == "text:  Hello, world!"
#     assert dash_duo.find_element("#data-entry-array")
#     assert dash_duo.find_element("#data-entry-array2")

#     assert dash_duo.get_logs() == [], "browser console should contain no error"

#     dash_duo.get_by


# def test_004_send_3D_array(dash_duo):
#     data_dashboard_app = DataDashboardApp(include_title=True)
#     dash_duo.start_server(data_dashboard_app.app)

#     dash_duo.wait_for_text_to_equal("#title", "Data Dashboard", timeout=4)
#     assert dash_duo.get_logs() == [], "browser console should contain no error"

#     send_data_to_dash(
#         {
#             "array": xr.DataArray(np.random.rand(10, 10, 10), name="my_arr_1"),
#         },
#         url="http://localhost:58050/data-dashboard",
#     )
#     sleep(1)
#     assert dash_duo.get_logs() == [], "browser console should contain no error"



import numpy as np
import plotly.express as px
import plotly.graph_objects as go
from plotly.graph_objects import Figure
from qua_dashboards.logging_config import logger
import xarray as xr
from typing import Any, Tuple, Union


def get_axis_info(data_array: xr.DataArray, dim: str) -> Tuple[Any, str]:
    """
    Retrieve the coordinate values and a formatted axis label (including units, if available)
    for the given dimension.

    Args:
        data_array: The xarray DataArray.
        dim: The dimension name.

    Returns:
        A tuple of (coordinate values, axis label).
    """
    if dim in data_array.coords:
        coord = data_array.coords[dim]
        units = coord.attrs.get("units") if hasattr(coord, "attrs") else None
    else:
        size = data_array.sizes.get(dim, len(data_array))
        coord = list(range(size))
        units = None
    label = dim + (f" ({units})" if units else "")
    return coord, label


def apply_log_transform(values: Union[np.ndarray, list], log_flag: bool) -> np.ndarray:
    """
    Apply a log10 transformation to the values if log_flag is True and all values are positive.
    Otherwise, return the original values.

    Args:
        values: The numeric values to potentially transform.
        log_flag: Whether to apply log transformation.

    Returns:
        The (possibly transformed) values as a numpy array.
    """
    arr = np.array(values)
    if log_flag:
        if (arr <= 0).any():
            logger.warning(
                "Non-positive values encountered for log scale; skipping log transform."
            )
            return arr
        return np.log10(arr)
    return arr


def update_axis_scaling(fig: Figure, data_array: xr.DataArray) -> None:
    """
    Update the x- and y-axis scaling on the figure based on log-scaling flags in data_array.attrs.

    Args:
        fig: The Plotly figure to update.
        data_array: The xarray DataArray containing potential log scale flags.
    """
    if data_array.attrs.get("log_x", False):
        fig.update_xaxes(type="log")
    if data_array.attrs.get("log_y", False):
        fig.update_yaxes(type="log")


def update_global_layout(fig: Figure, data_array: xr.DataArray) -> None:
    """
    Update the figure's layout with global settings from data_array.attrs.
    A title is applied only if explicitly provided in the attributes.

    Args:
        fig: The Plotly figure to update.
        data_array: The xarray DataArray containing global layout metadata.
    """
    layout = data_array.attrs.get("plot_layout", {})
    if "title" in data_array.attrs:
        fig.update_layout(title=data_array.attrs["title"], **layout)
    else:
        fig.update_layout(**layout)


def plot_1d(data_array: xr.DataArray) -> Figure:
    """
    Create a new 1D line plot for the given xarray DataArray.

    Args:
        data_array: A 1D xarray DataArray.

    Returns:
        A Plotly figure representing the line plot.
    """
    x_dim = data_array.dims[0]
    x, x_label = get_axis_info(data_array, x_dim)
    y_label = data_array.name if data_array.name is not None else "Value"
    if "units" in data_array.attrs:
        y_label += f" ({data_array.attrs['units']})"
    fig = px.line(x=x, y=data_array.values)
    fig.update_xaxes(title_text=x_label)
    fig.update_yaxes(title_text=y_label)
    update_axis_scaling(fig, data_array)
    update_global_layout(fig, data_array)
    return fig


def update_1d(fig: Figure, data_array: xr.DataArray) -> Figure:
    """
    Update an existing 1D line plot with new data from the xarray DataArray.

    Args:
        fig: The existing Plotly figure.
        data_array: A 1D xarray DataArray.

    Returns:
        The updated Plotly figure.
    """
    x_dim = data_array.dims[0]
    x, x_label = get_axis_info(data_array, x_dim)
    y_label = data_array.name if data_array.name is not None else "Value"
    if "units" in data_array.attrs:
        y_label += f" ({data_array.attrs['units']})"
    if not fig.data:
        fig.add_trace(go.Scatter())
    fig.update_traces(x=x, y=data_array.values)
    fig.update_xaxes(title_text=x_label)
    fig.update_yaxes(title_text=y_label)
    update_axis_scaling(fig, data_array)
    update_global_layout(fig, data_array)
    return fig


def plot_2d(data_array: xr.DataArray) -> Figure:
    """
    Create a new 2D heatmap plot for the given xarray DataArray.

    Args:
        data_array: A 2D xarray DataArray.

    Returns:
        A Plotly figure representing the heatmap.
    """
    y_dim, x_dim = data_array.dims[0], data_array.dims[1]
    x, x_label = get_axis_info(data_array, x_dim)
    y, y_label = get_axis_info(data_array, y_dim)
    z = apply_log_transform(data_array.values, data_array.attrs.get("log_z", False))
    # Build colorbar title from the data array's name and units.
    colorbar_title = ""
    if data_array.name:
        colorbar_title += data_array.name
    if "units" in data_array.attrs:
        colorbar_title += f" ({data_array.attrs['units']})"
    heatmap = go.Heatmap(z=z, x=x, y=y, colorbar=dict(title=colorbar_title))
    fig = go.Figure(data=heatmap)
    fig.update_xaxes(title_text=x_label)
    fig.update_yaxes(title_text=y_label)
    update_axis_scaling(fig, data_array)
    update_global_layout(fig, data_array)
    return fig


def update_2d(fig: Figure, data_array: xr.DataArray) -> Figure:
    """
    Update an existing 2D heatmap plot with new data from the xarray DataArray.

    Args:
        fig: The existing Plotly figure.
        data_array: A 2D xarray DataArray.

    Returns:
        The updated Plotly figure.
    """
    y_dim, x_dim = data_array.dims[0], data_array.dims[1]
    x, x_label = get_axis_info(data_array, x_dim)
    y, y_label = get_axis_info(data_array, y_dim)
    z = apply_log_transform(data_array.values, data_array.attrs.get("log_z", False))
    if not fig.data:
        fig.add_trace(go.Heatmap())
    fig.update_traces(z=z, x=x, y=y)
    fig.update_xaxes(title_text=x_label)
    fig.update_yaxes(title_text=y_label)
    update_axis_scaling(fig, data_array)
    # Update colorbar title for all heatmap traces.
    colorbar_title = ""
    if data_array.name:
        colorbar_title += data_array.name
    if "units" in data_array.attrs:
        colorbar_title += f" ({data_array.attrs['units']})"
    for trace in fig.data:
        if isinstance(trace, go.Heatmap):
            trace.colorbar.title = colorbar_title
    update_global_layout(fig, data_array)
    return fig


def plot_xarray(data_array: xr.DataArray, fig: Figure = None) -> Figure:
    """
    Create a new Plotly figure from an xarray DataArray.
    Supports 1D and 2D arrays.

    Args:
        data_array: An xarray DataArray (1D or 2D).
        fig: Optional existing figure (ignored in creation).

    Returns:
        A new Plotly figure.

    Raises:
        ValueError: If data_array.ndim is not 1 or 2.
    """
    logger.info(f"Plotting xarray data with dimensions: {data_array.dims}")
    if data_array.ndim == 1:
        return plot_1d(data_array)
    elif data_array.ndim == 2:
        return plot_2d(data_array)
    else:
        logger.error(f"Unsupported array dimensions: {data_array.ndim}")
        raise ValueError(
            f"Only 1D and 2D arrays are supported, not {data_array.ndim}D."
        )


def update_xarray_plot(fig: Figure, data_array: xr.DataArray) -> Figure:
    """
    Update an existing Plotly figure with new data from an xarray DataArray.
    Supports 1D and 2D arrays.

    Args:
        fig: The existing Plotly figure.
        data_array: An xarray DataArray (1D or 2D).

    Returns:
        The updated Plotly figure.

    Raises:
        ValueError: If data_array.ndim is not 1 or 2.
    """
    logger.info(f"Updating plot with xarray data: {data_array.name}")
    if data_array.ndim == 1:
        return update_1d(fig, data_array)
    elif data_array.ndim == 2:
        return update_2d(fig, data_array)
    else:
        logger.error(f"Unsupported array dimensions: {data_array.ndim}")
        raise ValueError(
            f"Only 1D and 2D arrays are supported, not {data_array.ndim}D."
        )



import dash_bootstrap_components as dbc
import dash
from dash import html, dcc, Input, Output, State, ALL, no_update
from dash.exceptions import PreventUpdate
import numpy as np
from qua_dashboards.core import BaseComponent
from .virtual_layer_adder import VirtualLayerAdder


class VirtualLayerEditor(BaseComponent):
    """
    A dash component allowing the editing of virtual layers to work with VirtualGateSet. Renders the matrix as an NxN editable table, complete with a reset_to_identity button. 
    An apply button changes the VirtualGateSet.layer matrix. 
    """


    def __init__(self,gateset,component_id = "VG_manager", dc_set = None):
        """
        Args:
            gateset: The VirtualGateSet instance
            component_id: Dash Component ID; also the name of the component title
        """
        super().__init__(component_id = component_id)
        self.gateset = gateset
        self.dc_set = dc_set
        self.component_id = component_id
        self.layer_names = [
            {"label": f"Layer {i+1}", "value": i}
            for i, _ in enumerate(self.gateset.layers)
        ]
        self._adder = VirtualLayerAdder(self.gateset, component_id = f"{component_id}-adder", layer_names=self.layer_names, dc_set=dc_set)

    @property
    def active_target(self): 
        """Returns gateset or dc_set based on current selection. Default is gateset"""
        return getattr(self, "_current_target_set", self.gateset)
    
    def _render_matrix_editor(self, layer_idx, use_dc = False):
        target_set = self.dc_set if (use_dc and self.dc_set) else self.gateset 
        if layer_idx >= len(target_set.layers): 
            return html.Div("Layer not available for this target")
        layer = target_set.layers[layer_idx]
        num_sources = len(layer.source_gates)
        num_targets = len(layer.target_gates)
        col_width = 12 // (num_targets + 1)
        header = [dbc.Col("", width=col_width)] + [
            dbc.Col(html.B(name), width=col_width) for name in layer.target_gates
        ]
        rows = []
        for i, row_name in enumerate(layer.source_gates):
            row = [dbc.Col(html.B(row_name), width=col_width)]
            for j in range(num_targets):
                row.append(
                    dbc.Col(
                        dcc.Input(
                            id={"type": "edit-matrix-cell", "layer": layer_idx, "row": i, "col": j},
                            type="number",
                            value=layer.matrix[i][j],
                            style={"width": "auto"}
                        ),
                        width=col_width,
                    )
                )
            rows.append(dbc.Row(row, className="mb-1"))
        return html.Div([dbc.Row(header, className="mb-2"), *rows])


    def get_layout(self):
        layers = [layer for layer in self.gateset.layers]
        num_of_layers = len(layers)
        options = self.layer_names
        default = 0 if num_of_layers > 0 else None

        opx_dc_toggle_switch = dbc.Row([
            dbc.Col(html.Label("OPX", className="me-2"), width="auto"),
            dbc.Col(
                dbc.Switch(
                    id=f"{self.component_id}-target-switch",
                    value=False,  # False = OPX, True = DC
                    className="mt-1",
                ),
                width="auto",
            ),
            dbc.Col(html.Label("DC"), width="auto"),
            dbc.Col(
                dbc.Checkbox(
                    id=f"{self.component_id}-edit-both",
                    label="Edit both",
                    value=False,
                    className="ms-4",
                ),
                width="auto",
            ),
        ], className="mb-3 align-items-center") if self.dc_set is not None else html.Div()

        editor_card = dbc.Card([
            dbc.CardHeader("VirtualGateSet Layer Editor"),
            dbc.CardBody([
                dcc.Dropdown(
                    id = f"{self.component_id}-layer-dropdown",
                    options = options,
                    value = default,
                    clearable = False
                ),
                html.Div(id = f"{self.component_id}-layer-matrix-editor"),
                dcc.Interval(
                    id=f"{self.component_id}-init",
                    interval=50, n_intervals=0, max_intervals=1
                ),
                dbc.Row(
                    [
                        dbc.Col(
                            dbc.Button(
                                "Apply Changes",
                                id=f"{self.component_id}-apply-edit",
                                color="primary",
                                className="mt-2 w-100",
                            )
                        ),
                        dbc.Col(
                            dbc.Button(
                                "Reset to Identity",
                                id=f"{self.component_id}-identity-reset",
                                color="secondary",
                                className="mt-2 w-100",
                            )
                        ),
                    ],
                    className="g-2",
                ),                
                html.Div(id=f"{self.component_id}-edit-output", style={"display": "none"})
            ])
        ])

        return html.Div([
            self._adder.get_layout(), 
            opx_dc_toggle_switch,
            editor_card, 
        ])


    def register_callbacks(self, app):
        self._adder.register_callbacks(app)

        @app.callback(
            Output(f"{self.component_id}-layer-matrix-editor", "children"),
            Output(f"{self.component_id}-layer-dropdown", "options"),
            Output(f"{self.component_id}-layer-dropdown", "value"),
            Input(f"{self.component_id}-layer-dropdown", "value"),
            Input(f"{self.component_id}-target-switch", "value") if self.dc_set else Input(f"{self.component_id}-init", "n_intervals"),
            Input("vg-layer-refresh-trigger", "data"),
            Input(f"{self.component_id}-init", "n_intervals"),
        )
        def show_selected_layer(layer_idx, use_dc, _refresh, _init):
            if self.dc_set is None:
                use_dc = False
            
            target_set = self.dc_set if (use_dc and self.dc_set) else self.gateset
            options = [{"label": f"Layer {i+1}", "value": i} for i in range(len(target_set.layers))]

            if not options:
                return "No Layers", [], None
            if layer_idx is None or layer_idx >= len(target_set.layers):
                layer_idx = 0

            return self._render_matrix_editor(layer_idx, use_dc), options, layer_idx


        @app.callback(
            Output(f"{self.component_id}-edit-output", "children"),
            Output("vg-layer-refresh-trigger", "data", allow_duplicate=True),
            Input(f"{self.component_id}-apply-edit", "n_clicks"),
            State(f"{self.component_id}-layer-dropdown", "value"), 
            State(f"{self.component_id}-target-switch", "value") if self.dc_set else State(f"{self.component_id}-layer-dropdown", "value"),
            State(f"{self.component_id}-edit-both", "value") if self.dc_set else State(f"{self.component_id}-layer-dropdown", "value"),
            State({'type': 'edit-matrix-cell', 'layer': ALL, 'row': ALL, 'col': ALL}, 'value'),
            State({'type': 'edit-matrix-cell', 'layer': ALL, 'row': ALL, 'col': ALL}, 'id'),
            State("vg-layer-refresh-trigger", "data"),
            prevent_initial_call=True
        )
        def apply_layer_changes(n_clicks, layer_idx, use_dc, edit_both, matrix_flat, cell_ids, vg_val):
            if not n_clicks:
                raise dash.exceptions.PreventUpdate
            if self.dc_set is None:
                edit_both = False
                use_dc = False

            target_set = self.dc_set if (use_dc and self.dc_set) else self.gateset
            layer = target_set.layers[layer_idx]
            num_sources = len(layer.source_gates)
            num_targets = len(layer.target_gates)
            coord_vals = {(cid["row"], cid["col"]): v for cid, v in zip(cell_ids, matrix_flat)}
            M = [
                [float(coord_vals.get((i, j), 0.0) or 0.0) for j in range(num_targets)]
                for i in range(num_sources)
            ]
            if edit_both:
                if layer_idx < len(self.gateset.layers):
                    self.gateset.layers[layer_idx].matrix = [row[:] for row in M]
                if self.dc_set is not None and layer_idx < len(self.dc_set.layers):
                    self.dc_set.layers[layer_idx].matrix = [row[:] for row in M]
            else:
                target_set.layers[layer_idx].matrix = M
            return "Updated!", (vg_val or 0) + 1
        
        @app.callback(
            Output(f"{self.component_id}-layer-matrix-editor", "children", allow_duplicate=True),
            Output("vg-layer-refresh-trigger", "data", allow_duplicate=True),
            Input(f"{self.component_id}-identity-reset", 'n_clicks'),
            State(f"{self.component_id}-layer-dropdown", 'value'),
            State(f"{self.component_id}-target-switch", "value") if self.dc_set else State(f"{self.component_id}-layer-dropdown", "value"),
            State(f"{self.component_id}-edit-both", "value") if self.dc_set else State(f"{self.component_id}-layer-dropdown", "value"),
            State("vg-layer-refresh-trigger", "data"),
            prevent_initial_call=True
        )
        def reset_to_identity_and_refresh(n_clicks, layer_idx, use_dc, edit_both, vg_val):
            if not n_clicks or layer_idx is None:
                raise PreventUpdate
            
            if self.dc_set is None:
                edit_both = False
                use_dc = False
            
            target_set = self.dc_set if (use_dc and self.dc_set) else self.gateset
            layer = target_set.layers[layer_idx]
            num_sources = len(layer.source_gates)
            num_targets = len(layer.target_gates)
            identity = [[1.0 if i == j else 0.0 for j in range(num_targets)] for i in range(num_sources)]
            if edit_both:
                if layer_idx < len(self.gateset.layers):
                    self.gateset.layers[layer_idx].matrix = [row[:] for row in identity]
                if layer_idx < len(self.dc_set.layers):
                    self.dc_set.layers[layer_idx].matrix = [row[:] for row in identity]
            else:
                target_set.layers[layer_idx].matrix = identity
            
            return self._render_matrix_editor(layer_idx, use_dc), (vg_val or 0) + 1



import logging
from typing import Any, Dict, List, Optional
import plotly.graph_objects as go
from dash import Dash, Input, Output, State, dcc, html

from qua_dashboards.core import BaseComponent
from qua_dashboards.video_mode import data_registry
from qua_dashboards.video_mode.utils.plot_utils import *


logger = logging.getLogger(__name__)

__all__ = ["SharedViewerComponent"]


class SharedViewerComponent(BaseComponent):
    """
    A component responsible for displaying data visualizations.

    It listens to a primary data store for references to data in a
    server-side registry (which can be live data or a static compound object
    containing a base image and annotations) and a layout configuration store.
    """

    _MAIN_GRAPH_ID_SUFFIX = "main-graph"

    def __init__(self, component_id: str, **kwargs: Any) -> None:
        """Initializes the SharedViewerComponent.

        Args:
            component_id: A unique string identifier for this component instance.
            **kwargs: Additional keyword arguments passed to BaseComponent.
        """
        super().__init__(component_id=component_id, **kwargs)
        self._current_figure: go.Figure = (
            self._get_default_figure()
        )  # Start with an empty dark figure
        self._last_layout_update: Optional[Dict] = None
        self._last_layout_shapes: List[Dict] = []
        logger.info(f"SharedViewerComponent '{self.component_id}' initialized.")

    def _get_default_figure(self) -> go.Figure:
        """
        Returns a default empty Plotly figure with the dark template.
        """
        return go.Figure().update_layout(template="plotly_dark")

    def get_layout(self) -> html.Div:
        """Generates the Dash layout for the SharedViewerComponent.

        Returns:
            An html.Div component containing the dcc.Graph.
        """
        return html.Div(
            style={"height": "100%", "width": "100%"},
            children=[
                dcc.Graph(
                    id=self._get_id(self._MAIN_GRAPH_ID_SUFFIX),
                    figure=self._current_figure,
                    style={"height": "100%", "width": "100%"},
                    config={"scrollZoom": True, "displaylogo": False, "displayModeBar": False},
                )
            ],
        )

    def get_graph_id(self) -> Dict[str, str]:
        """
        Returns the Dash ID dictionary for the main graph component.

        This ID is needed by other components (e.g., AnnotationTabController)
        to register callbacks for graph interactions.

        Returns:
            A dictionary suitable for use as a Dash component ID.
        """
        return self._get_id(self._MAIN_GRAPH_ID_SUFFIX)

    def _create_figure_from_live_data(self, data_object: dict) -> go.Figure:
        """
        Creates a Plotly figure from live data (expected to be xr.DataArray).

        Args:
            data_object: The data object fetched from the registry.

        Returns:
            A Plotly Figure object.
        """
        try:
            fig = build_live_figure(data_object)
        except Exception as e:
            logger.error(f"SharedViewer ({self.component_id}): live->fig error: {e}", exc_info=True)
            fig = empty_dark()
        self._current_figure = fig
        return fig


    def _create_figure_from_static_data(self, static_data_object: dict, viewer_ui_state_input: dict) -> go.Figure:
        """
        Creates a Plotly figure from a static data compound object.

        The compound object is expected to contain 'base_image_data'
        and 'annotations'.

        Args:
            static_data_object: The compound static data object from the registry.

        Returns:
            A Plotly Figure object with base image and annotations.
        """
        try:
            fig = build_static_figure(static_data_object, viewer_ui_state_input)
        except Exception as e:
            logger.error(f"SharedViewer ({self.component_id}): static->fig error: {e}", exc_info=True)
            fig = empty_dark()
        self._current_figure = fig
        return fig


    def register_callbacks(
        self,
        app: Dash,
        viewer_data_store_id: Dict[str, str],
        viewer_ui_state_store_id: Dict[str, Any],
        layout_config_store_id: Dict[str, str],
    ) -> None:
        """Registers Dash callbacks for the SharedViewerComponent.

        Args:
            app: The main Dash application instance.
            viewer_data_store_id: The ID of the store containing the reference
                to the primary data (key and version).
            layout_config_store_id: The ID of the store containing layout updates.
        """
        logger.debug(
            f"Registering callbacks for SharedViewerComponent '{self.component_id}'"
        )

        @app.callback(
            Output(self._get_id(self._MAIN_GRAPH_ID_SUFFIX), "figure"),
            Input(viewer_data_store_id, "data"),
            Input(viewer_ui_state_store_id, "data"),
            Input(layout_config_store_id, "data"),
            State(self._get_id(self._MAIN_GRAPH_ID_SUFFIX), "figure"),
            prevent_initial_call=True,  # Prevent update on initial app load if stores are empty
        )
        def update_shared_viewer_graph(
            viewer_data_ref: Optional[Dict[str, Any]],
            viewer_ui_state_input: Optional[Dict[str, Any]],
            layout_updates_input: Optional[Dict[str, Any]],
            current_fig_state_dict: Optional[Dict[str, Any]],
        ) -> go.Figure:
            fig_to_display = self._get_default_figure()  # Default to empty dark figure

            if viewer_data_ref is None or not isinstance(viewer_data_ref, dict):
                logger.debug(
                    f"SharedViewer ({self.component_id}): viewer_data_ref is None or not a dict. "
                    "No primary data to display."
                )
                # Apply layout updates even to an empty figure if provided
                if isinstance(layout_updates_input, dict):
                    fig_to_display.update_layout(layout_updates_input)
                self._current_figure = fig_to_display  # Store the figure
                return fig_to_display

            data_key = viewer_data_ref.get("key")
            # version = viewer_data_ref.get("version") # Version can be used for debugging

            if not data_key:
                logger.warning(
                    f"SharedViewer ({self.component_id}): viewer_data_ref missing 'key'."
                )
            else:
                data_object = data_registry.get_data(data_key)
                if data_object is None:
                    logger.warning(
                        f"SharedViewer ({self.component_id}): Data not found in registry "
                        f"for key '{data_key}'. Displaying empty graph."
                    )
                elif data_key == data_registry.LIVE_DATA_KEY:
                    # logger.debug(
                    #     f"SharedViewer ({self.component_id}): Processing live data "
                    #     f"for key '{data_key}'."
                    # )
                    fig_to_display = self._create_figure_from_live_data(data_object)
                elif data_key == data_registry.STATIC_DATA_KEY:
                    logger.debug(
                        f"SharedViewer ({self.component_id}): Processing static data "
                        f"for key '{data_key}'."
                    )
                    fig_to_display = self._create_figure_from_static_data(data_object,viewer_ui_state_input)
                    fig_to_display.update_layout(showlegend=False)
                else:
                    logger.warning(
                        f"SharedViewer ({self.component_id}): Unrecognized data key "
                        f"'{data_key}'. Displaying empty graph."
                    )

            # Apply layout updates from the dedicated store
            if isinstance(layout_updates_input, dict):
                if "shapes" in layout_updates_input:
                    self._last_layout_shapes = layout_updates_input["shapes"]
                non_shape_updates = {k: v for k, v in layout_updates_input.items() if k != "shapes"}
                if non_shape_updates:
                    logger.debug(
                        f"SharedViewer ({self.component_id}): Applying layout updates: "
                        f"{layout_updates_input}"
                    )
                    fig_to_display.update_layout(non_shape_updates)
                # if layout_updates_input != self._last_layout_update:
                #     fig_to_display.update_layout(layout_updates_input)
                #     self._last_layout_update = layout_updates_input
            if self._last_layout_shapes:
                fig_to_display.update_layout(shapes=self._last_layout_shapes)

            # Commented out because it causes annotation points to become barely visible
            # Preserve zoom/pan state (uirevision)
            # If layout_updates_input itself contains 'uirevision', that will take precedence.
            # Otherwise, we try to preserve the existing uirevision.
            # if not (
            #     isinstance(layout_updates_input, dict)
            #     and "uirevision" in layout_updates_input
            # ):
            #     previous_uirevision = True  # Default to True to preserve state
            #     if current_fig_state_dict and isinstance(
            #         current_fig_state_dict.get("layout"), dict
            #     ):
            #         previous_uirevision = current_fig_state_dict["layout"].get(
            #             "uirevision", True
            #         )
            #     fig_to_display.layout.uirevision = previous_uirevision
            self._current_figure = fig_to_display  # Store the figure
            return fig_to_display

    def get_figures(self) -> List[go.Figure]:
        """
        Returns a list containing the current figure managed by this component.

        This figure is the last one that was processed and output by the
        component's update callback.

        Returns:
            A list containing a single plotly.graph_objects.Figure.
        """
        return [self._current_figure]


"""
Dash component for displaying and controlling a series of voltage parameters.
"""

import logging
from typing import Dict, Sequence

import dash
from dash import (
    dcc,
    html,
    Input,
    Output,
    ALL, 
    MATCH,
    State
)

from qua_dashboards.core import BaseComponent, ParameterProtocol

from .voltage_control_row import VoltageControlRow, format_voltage

from quam_builder.architecture.quantum_dots.components.virtual_dc_set import VirtualDCSet

logger = logging.getLogger(__name__)

DEFAULT_INPUT_CLASS_NAME = ""  # Input fields will always have this class
COMPONENT_MAX_WIDTH = "450px"


class VirtualGateParameter:
    """
    A Parameter to be integrated into the VoltageControlComponent, integrating the layers of virtual gates into the VoltageControlComponent. 
    """
    def __init__(
        self, 
        name: str, 
        label: str, 
        dc_set: VirtualDCSet, 
        unit: str = "V",
    ): 
        self.name = name
        self.dc_set = dc_set
        self.label = label
        self.unit = unit

    def get_latest(self): 
        return self.dc_set.get_voltage(self.name)
    
    def set(self, value) -> None:
        self.dc_set.set_voltages({self.name: value}, requery = False, resync = False)


class VoltageControlComponent(BaseComponent):
    """
    A Dash component to display and set voltages from multiple sources.
    Uses VoltageControlRow to manage individual parameter rows.
    Simplified: No edit highlighting, no auto-defocus on enter.
    """

    def __init__(
        self,
        component_id: str,
        voltage_parameters: Sequence[ParameterProtocol] = None,
        dc_set: VirtualDCSet = None,
        update_interval_ms: int = 1000,
        layout_columns: int = 3,
        step_size: float = 100e-6,
        preselected_gates: Sequence[str] | None = None,
    ):
        super().__init__(component_id=component_id)
        if voltage_parameters is None and dc_set is None: 
            raise ValueError("Please provide either a sequence of VoltageParameter objects, or a VirtualDCSet.")
        if voltage_parameters is None: 
            self.dc_set = dc_set
            _ = self.dc_set.all_current_voltages
            voltage_parameters = []
            for ch in self.dc_set.valid_channel_names: 
                virtual_param = VirtualGateParameter(name = ch, label = ch, dc_set = self.dc_set)
                voltage_parameters.append(virtual_param)
        self.update_interval_ms = update_interval_ms
        self.voltage_parameters = voltage_parameters
        self._initial_values_loaded = False  # To ensure first update populates values

        self._row_components: Dict[str, VoltageControlRow] = {}
        input_id_type = self._get_id_type_str("input")
        for param in self.voltage_parameters:
            row = VoltageControlRow(input_id_type=input_id_type, param=param)
            self._row_components[param.name] = row

        self.layout_columns = layout_columns
        self.step_size = step_size

        # Full list of available parameters is stored in self.voltage_parameters, and selected ones are appended to gates_to_display
        all_names = {p.name for p in self.voltage_parameters}
        self._initial_selected_names = [n for n in (preselected_gates or []) if n in all_names]
        self.gates_to_display = [p for p in self.voltage_parameters if p.name in self._initial_selected_names]

    def _get_id_type_str(self, element_name: str) -> str:
        return f"comp-{self.component_id}-{element_name}"
        
    def _build_dropdown_options(self) -> list:
        """Build dropdown options with physical/virtual grouping."""
        options = []
        physical_names = set(self.dc_set.channels.keys()) if self.dc_set else set()
        virtual_names = [n for n in self.dc_set.valid_channel_names if n not in physical_names]
        
        # Physical gates section
        if physical_names:
            options.append({"label": "── Physical Gates ──", "value": "__physical_header__", "disabled": True})
            for name in sorted(physical_names):
                options.append({"label": name, "value": name})
        
        # Virtual gates section  
        if virtual_names:
            options.append({"label": "── Virtual Gates ──", "value": "__virtual_header__", "disabled": True})
            for name in virtual_names:
                options.append({"label": name, "value": name})
        
        return options
    
    @property
    def voltage_parameters_by_name(self) -> Dict[str, "VirtualGateParameter"]:
        # Initial lazy create the param, which is then REused by the update_displayed_rows
        if self.dc_set is not None: 
            existing_names = {p.name for p in self.voltage_parameters}
            for name in self.dc_set.valid_channel_names: 
                if name not in existing_names: 
                    param = VirtualGateParameter(name = name, label = name, dc_set = self.dc_set)
                    self.voltage_parameters.append(param)

        return {p.name: p for p in self.voltage_parameters}

    def get_layout(self) -> html.Div:
        if not self._row_components:
            return html.Div(
                "No voltage parameters configured.",
                style={"maxWidth": COMPONENT_MAX_WIDTH},
            )
        input_ids = [
            {"type": self._get_id_type_str("input"), "index": param.name} 
            for param in self.gates_to_display
        ]
        return html.Div(
            [
                dcc.Dropdown(
                    id = self._get_id("dc-gate-selector"), 
                    options = self._build_dropdown_options(),
                    value = self._initial_selected_names, 
                    multi = True, 
                    placeholder = "Select gates to display", 
                    style = {"color": "black"}
                ),
                html.Div(
                    id = self._get_id("rows-container"), 
                    children = [self._row_components[n].get_layout() for n in self._initial_selected_names],
                ),
                dcc.Interval(
                    id=self._get_id("update-interval"),
                    interval=self.update_interval_ms,
                    n_intervals=0,
                ),
                dcc.Store(
                    id=self._get_id("keyboard-config"),
                    data={
                        "input_ids": input_ids,
                        "step_size": self.step_size,
                        "dummy_store_id": self._get_id("keyboard-dummy"),
                    }
                ),
                dcc.Store(id=self._get_id("keyboard-dummy")),  # Changed from html.Div
                # *[row.get_layout() for row in self._row_components.values()],
            ],
            style={"maxWidth": COMPONENT_MAX_WIDTH, "margin": "0 auto"},
        )
    
    def register_callbacks(self, app: dash.Dash) -> None:
        self._app = app
        if not self.voltage_parameters:
            return

        input_id_type_str = self._get_id_type_str("input")

        @app.callback(
            Output({"type": input_id_type_str, "index": ALL}, "value", allow_duplicate=True), 
            Output({"type": input_id_type_str, "index": ALL}, "className", allow_duplicate=True), 
            Input(self._get_id("update-interval"), "n_intervals"), 
            State({"type": input_id_type_str, "index": ALL}, "id"),
            prevent_initial_call = True, 
        )
        def periodic_update(_n_intervals, input_ids): 
            if not input_ids: 
                return [],[]
            values = []
            classnames = []
            for input_id in input_ids: 

                param_name = input_id["index"]
                control_row = self._row_components[param_name]

                param = next(p for p in self.voltage_parameters if p.name == param_name)

                try:
                    live_value = param.get_latest()
                    live_text = format_voltage(live_value)
                except Exception as e:
                    logger.error(f"Err get_latest for {param_name}: {e}", exc_info=True)
                    values.append(dash.no_update)
                    classnames.append(dash.no_update)
                    continue

                if control_row.current_input_text != control_row.last_committed_text:
                    values.append(dash.no_update)
                elif live_text != control_row.current_input_text:
                    control_row.current_input_text = live_text
                    control_row.last_committed_text = live_text
                    values.append(live_text)
                else: 
                    values.append(dash.no_update)
                classnames.append(DEFAULT_INPUT_CLASS_NAME)
            return values, classnames
        
        @app.callback(
            Output(self._get_id("rows-container"), "children"), 
            Output(self._get_id("keyboard-config"), "data"),
            Input(self._get_id("dc-gate-selector"), "value"),
        )
        def update_displayed_rows(selected_names): 
            if not selected_names: 
                return [], {"input_ids": [], "step_size": self.step_size}
            params_by_name = self.voltage_parameters_by_name
            for name in selected_names:
                if name not in self._row_components:
                    param = params_by_name[name]
                    # param = VirtualGateParameter(name=name, label=name, dc_set=self.dc_set)
                    # self.voltage_parameters.append(param)
                    row = VoltageControlRow(input_id_type=self._get_id_type_str("input"), param=param)
                    self._row_components[name] = row
            self.gates_to_display = [p for p in self.voltage_parameters if p.name in selected_names]
            rows = [self._row_components[name].get_layout() for name in selected_names]
            input_ids = [{"type": self._get_id_type_str("input"), "index": name} for name in selected_names]

            return rows, {"input_ids": input_ids, "step_size": self.step_size, "dummy_store_id": self._get_id("keyboard-dummy")}
        
        @app.callback(
            Output({"type": input_id_type_str, "index": MATCH}, "value", allow_duplicate=True), 
            Input({"type": input_id_type_str, "index": MATCH}, "n_submit"), 
            State({"type": input_id_type_str, "index": MATCH}, "value"), 
            State({"type": input_id_type_str, "index": MATCH}, "id"),
            prevent_initial_call = True, 
        )

        def handle_input_submit(_n_submit, submitted_text_value, input_id): 
            if not _n_submit or submitted_text_value is None:
                raise dash.exceptions.PreventUpdate
            
            param_name = input_id["index"]
            control_row = self._row_components.get(param_name)
            param = next((p for p in self.voltage_parameters if p.name == param_name), None)
            
            if control_row is None or param is None:
                raise dash.exceptions.PreventUpdate
            
            try:
                float_value = float(submitted_text_value)
                param.set(float_value)
                control_row.current_input_text = format_voltage(float_value)
                control_row.last_committed_text = control_row.current_input_text
                return control_row.current_input_text
            except ValueError:
                logger.warning(f"Invalid float: '{submitted_text_value}' for {param_name}")
                return dash.no_update
            except Exception as e:
                logger.error(f"Error during set for {param_name}: {e}", exc_info=True)
                return dash.no_update
            
        @app.callback(
            Output({"type": input_id_type_str, "index": ALL}, "value", allow_duplicate=True),
            Input(self._get_id("keyboard-dummy"), "data"),
            State({"type": input_id_type_str, "index": ALL}, "id"),
            prevent_initial_call=True,
        )
        def handle_arrow_key_change(store_data, input_ids):
            if not store_data or "param_name" not in store_data:
                raise dash.exceptions.PreventUpdate
            
            param_name = store_data["param_name"]
            new_value_str = store_data["value"]
            
            param = next((p for p in self.voltage_parameters if p.name == param_name), None)
            control_row = self._row_components.get(param_name)
            
            if param is None or control_row is None:
                raise dash.exceptions.PreventUpdate
            
            try:
                float_value = float(new_value_str)
                param.set(float_value)
                formatted = format_voltage(float_value)
                control_row.current_input_text = formatted
                control_row.last_committed_text = formatted
            except (ValueError, Exception) as e:
                logger.error(f"Arrow key set error for {param_name}: {e}")
                raise dash.exceptions.PreventUpdate
            
            # Return updated value for the matching input, no_update for the rest
            return [
                formatted if iid["index"] == param_name else dash.no_update
                for iid in input_ids
            ]

        @app.callback(
            Output({"type": input_id_type_str, "index": MATCH}, "value", allow_duplicate=True),
            Input({"type": input_id_type_str, "index": MATCH}, "n_blur"),
            State({"type": input_id_type_str, "index": MATCH}, "id"),
            prevent_initial_call=True,
        )
        def handle_input_blur(_n_blur, input_id):
            if not _n_blur:
                raise dash.exceptions.PreventUpdate
            
            param_name = input_id["index"]
            control_row = self._row_components.get(param_name)
            param = next((p for p in self.voltage_parameters if p.name == param_name), None)
            
            if control_row is None or param is None:
                raise dash.exceptions.PreventUpdate
            
            control_row.current_input_text = format_voltage(param.get_latest())
            control_row.last_committed_text = control_row.current_input_text
            return control_row.current_input_text

        app.clientside_callback(
            r"""
            function(config) {
                if (!config || !config.input_ids) {
                    return null;
                }

                if (window._voltageKeyboardListenerAttached) {
                    return null;
                }
                window._voltageKeyboardListenerAttached = true;

                var inputIds = config.input_ids || [];
                var stepSize = config.step_size || 0.00001;

                function findVoltageInput(index) {
                    if (index < 0 || index >= inputIds.length) {
                        return null;
                    }
                    var id = inputIds[index];
                    var allInputs = document.querySelectorAll('input');
                    for (var i = 0; i < allInputs.length; i++) {
                        var el = allInputs[i];
                        var rawId = el.id;
                        if (!rawId) continue;
                        try {
                            var parsed = JSON.parse(rawId);
                            if (parsed && parsed.type === id.type && parsed.index === id.index) {
                                return el;
                            }
                        } catch (e) {}
                    }
                    return null;
                }

                function isVoltageInputElement(el) {
                    if (!el || !el.id) return false;
                    try {
                        var parsed = JSON.parse(el.id);
                        for (var i = 0; i < inputIds.length; i++) {
                            var id = inputIds[i];
                            if (id.type === parsed.type && id.index === parsed.index) {
                                return true;
                            }
                        }
                        return false;
                    } catch (e) {
                        return false;
                    }
                }

                document.addEventListener('keydown', function(e) {
                    var activeEl = document.activeElement;
                    var tagName = activeEl ? activeEl.tagName.toLowerCase() : '';
                    var isInInput = (tagName === 'input' || tagName === 'textarea');
                    var isInVoltageInput = isInInput && isVoltageInputElement(activeEl);

                    var shortcuts = [
                        'q','w','e','r','t','y','u','i','o','p',
                        'a','s','d','f','g','h','j','k','l',
                        'z','x','c','v','b','n','m'
                    ];
                    var key = (e.key || '').toLowerCase();
                    var idx = shortcuts.indexOf(key);

                    if (idx !== -1 && (!isInInput || isInVoltageInput)) {
                        var targetEl = findVoltageInput(idx);
                        if (targetEl) {
                            e.preventDefault();
                            if (activeEl && typeof activeEl.blur === 'function') {
                                activeEl.blur();
                            }
                            targetEl.focus();
                            if (typeof targetEl.select === 'function') {
                                targetEl.select();
                            }
                        }
                    }

                    if (isInVoltageInput && (e.key === 'ArrowUp' || e.key === 'ArrowDown')) {
                        e.preventDefault();

                        var currentVal = parseFloat(activeEl.value);
                        if (!isFinite(currentVal)) {
                            currentVal = 0;
                        }

                        var delta = (e.key === 'ArrowUp') ? stepSize : -stepSize;
                        if (e.shiftKey) delta *= 10;
                        if (e.ctrlKey)  delta *= 100;

                        var newVal = currentVal + delta;
                        var newValStr = newVal.toFixed(9).replace(/\.?0+$/, '');

                        activeEl.value = newValStr;
                        var parsedId = JSON.parse(activeEl.id);
                        dash_clientside.set_props(
                            config.dummy_store_id,
                            {data: {param_name: parsedId.index, value: newValStr, ts: Date.now()}}
                        );
                    }
                });

                return null;
            }
            """,
            Output(self._get_id("keyboard-dummy"), "data"),
            Input(self._get_id("keyboard-config"), "data"),
            prevent_initial_call=True,
        )



from abc import ABC
from enum import Flag, auto

from qua_dashboards.core.base_component import BaseComponent

from dash import html
from typing import Any, Dict, List


__all__ = ["ModifiedFlags", "BaseUpdatableComponent"]


class ModifiedFlags(Flag):
    """Flags indicating what needs to be modified after parameter changes."""

    NONE = 0
    PARAMETERS_MODIFIED = auto()
    PROGRAM_MODIFIED = auto()
    CONFIG_MODIFIED = auto()


class BaseUpdatableComponent(BaseComponent, ABC):
    """
    Extends BaseComponent for components with UI-updatable parameters.

    This class provides a standard interface for components whose parameters
    can be modified through the Dash UI, allowing for dynamic updates and
    communication of what aspects of the component were affected by these
    changes (e.g., if a program needs regeneration or a configuration
    needs to be reloaded).
    """

    def __init__(self, *args: Any, component_id: str, **kwargs: Any) -> None:
        """
        Initializes the BaseUpdatableComponent.

        Args:
            component_id: A unique string identifier for this component instance.
        """
        # Pass through component_id and any other args/kwargs to BaseComponent
        super().__init__(component_id=component_id, *args, **kwargs)  # type: ignore

    def update_parameters(self, parameters: Dict[str, Dict[str, Any]]) -> ModifiedFlags:
        """
        Update the component's attributes based on the input values from the UI.

        Subclasses should implement this method to handle changes to their
        specific parameters. The `parameters` dictionary is typically structured
        with the component's own ID as a key.

        Args:
            parameters: A dictionary where keys are component IDs (matching
                `self.component_id` for this component's parameters) and values are
                dictionaries of parameter names to their new values.
                Example: {'my-component-id': {'param_a': 10, 'param_b': 'value'}}

        Returns:
            ModifiedFlags indicating what aspects of the component were changed
            and might require further action (e.g., re-generating a program).
            Defaults to ModifiedFlags.NONE.
        """
        return ModifiedFlags.NONE

    def get_dash_components(self, include_subcomponents: bool = True) -> List[html.Div]:
        """
        Return a list of Dash components representing the UI for editing parameters.

        Subclasses should override this to provide the specific input fields
        or controls for their manageable parameters.

        Args:
            include_subcomponents (bool, optional): Whether to include UI components
                from sub-components that this component might manage. Defaults to True.

        Returns:
            A list of Dash html.Div or similar components. Defaults to an empty list.
        """
        return []

    def get_components(self) -> List["BaseUpdatableComponent"]:
        """
        Return a list of component instances for this component's parameter UIs.

        These components are used in Dash callbacks to manage and update parameter
        values. Typically, this will be `[self]`, but can include managed sub-components
        if `include_subcomponents` was True in `get_dash_components`.

        Returns:
            A list of component instances. Defaults to `[self]`.
        """
        return [self]
